.row-content {
  padding: 10px;
}

.global-toggle-row {
  padding: 4px;
}

.content {
  margin: 24px 12px;
}

.detail {
  margin: 24px;
}

.multiline {
  line-height: 140%;
}

.tint {
	background-color: rgba(0,0,0,0.8);
}

.search-row {
  padding: 10px 5px;
}

.version-label {
  background-color: #62a0ea;
  color: white;
  border-radius: 4px;
  padding: 4px;
  box-shadow: 0 0 2px rgba(0,0,0,0.2);
}

.update {
  color: #ff7800;
}

.update-bar {
  border-top: 1px solid @borders;
  padding: 8px 12px;
}

.comment-tile {
  padding: 16px;
  padding-bottom: 24px;
}

.detail-heading {
  margin-bottom: 6px;
}

rating {
  color: @yellow_4;
}

.upgrade-assistant-result {
  padding: 12px;
}

/* Progress Bar Colours */
progressbar.error progress {
    background-color: @red_1;
}

progressbar.error text {
    color: @red_1;
    opacity: 0.8;
}

progressbar.warning progress {
    background-color: @yellow_4;
}

progressbar.warning text {
    color: @window_fg_color;
}

progressbar.success progress {
    background-color: @green_1;
}

progressbar.success text {
    color: @green_1;
    opacity: 0.8;
}
